import numpy as np
import pandas as pd

def get_nearest_value(sim_values,value):
    """
        Given a value and a values list this method return from the the nearest element to the value
    """
    absolute_val_array = np.abs(sim_values - value)
    return sim_values[absolute_val_array.argmin()]

def block_dataset(raw_dfa,raw_dfb,blk_att_a,blk_att_b,
                    id_a='voter_id',id_b='voter_id',
                    key_lenght=3):
    """
        Block a dataset considering one attributes
        - raw_dfa : dataset a
        - raw_dfa : dataset b
        - blk_att_a : attribute to block
        - blk_att_b : attribute to block
        - id_a : id attribute (default: voter_id)
        - id_b : id attribute (default: voter_id)
        - key_lenght : key kength
    """

    if type(raw_dfa) == list:
      raw_dfa = pd.DataFrame(raw_dfa)
      raw_dfa.columns = ['id','val']
      raw_dfb = pd.DataFrame(raw_dfb)
      raw_dfb.columns = ['id','val']

      blk_att_a = 'val'
      blk_att_b = 'val'
      id_a = 'id'
      id_b = 'id'

    raw_dfa['BLK'] = raw_dfa[blk_att_a].str[:key_lenght]
    raw_dfb['BLK'] = raw_dfb[blk_att_b].str[:key_lenght]

    comps = []

    for block in raw_dfa['BLK'].unique():
        fdfa = raw_dfa[raw_dfa['BLK']==block]
        fdfb = raw_dfb[raw_dfb['BLK']==block]

        if len(fdfb) > 0:
            for ida in fdfa[id_a].unique():
                for idb in fdfb[id_b].unique():
                    comps.append([ida,idb]) #adiciona a comparacao
    
    comps = pd.DataFrame(comps)
    comps.columns = ['id1','id2']
    return comps


import unittest
from ds_util.bloomfilter import create_bf, jaccard_similarity

class AnonimizationTest(unittest.TestCase):

    def test_bf_creation(self):
        with self.assertRaises(AssertionError):
            create_bf('abc',20,3,bf_representation='none')

    def test_jaccard_sim(self):
        a='ANA'
        b='ANE'

        bfa_b = create_bf(a,20,3,bf_representation='binary')
        bfb_b = create_bf(b,20,3,bf_representation='binary')

        bfa_pos1 = create_bf(a,20,3,bf_representation='pos1')
        bfb_pos1 = create_bf(b,20,3,bf_representation='pos1')

        sim_b = jaccard_similarity(bfa_b,bfb_b,bf_representation='binary')
        sim_pos =  jaccard_similarity(bfa_pos1,bfb_pos1,bf_representation='pos1')

        self.assertAlmostEqual(sim_b,sim_pos)
        self.assertEqual(sim_b,sim_pos)

        self.assertGreater(jaccard_similarity(bfa_b,bfb_b,bf_representation='binary'),.4)
        

if __name__ == '__main__':
    unittest.main()


from multiprocessing import Pool

import numpy as np
import pandas as pd

import nltk

from ds_util.util import get_nearest_value
from ds_util.bloomfilter import create_bf, jaccard_similarity, fill_pos_bf


def anonymize_dataset(ldf,bf_len,k,bf_representation='binary',num_ngrams=2):
  """
    Aanonymize a dataset with a bloomfilter. 

    - ldf: dataset
    - bf_len: bloomfilter ength
    - k : number of hashfunction 
    - bf_representation [ binary, pos1 ]

  """
  saida =[]
  
  for row in ldf:
    id = row[0]
    bf = create_bf(row[1],bf_len,k,bf_representation=bf_representation,num_ngrams=num_ngrams)
    saida.append([id,bf])
  
  return saida

# TODO: Implement
# def parallel_anonymize_dataset(df,bf_len,k,bf_representation='binary',cores=2):
#   """
#     Aanonymize a dataset with a bloomfilter. 

#     - ldf: dataset
#     - bf_len: bloomfilter ength
#     - k : number of hashfunction 
#     - bf_representation [ binary, pos1 ]

#   """
#   nparts_df = np.array_split(df, cores)
  
#   params = []
#   for dfa in nparts_df:
#       params.append((dfa,bf_len,k)) #kargs

#   with Pool(processes=cores) as p:
#       r = p.map(anonymize_dataset,params)

#   return r

def rank_dataset(df_a,df_b,bf_representation='binary',bit_power=1,blk=[]):

  '''
    Sort the dataset according to the jaccard simillarity. This method also change
    binary bf represation. It convert the bitaary into np.array(uint8) to be used 
    as input to the classifiers
    
    - bit_power: (default=1), this method is used amplify the power of the signal in the bloomfilter
    - blk : considering a block strategy if blk is [] no blk strategie will be employed (default [])

    return a dataset with ["id1","id2","bf1","bf2","sim","label"]
  '''
  
  if len(blk) == 0:
    return __rank_all_dataset(df_a,df_b,bf_representation='binary',bit_power=1)
  else:
    ds = []
    #converte para DataFrame para facilitar
    a = pd.DataFrame(df_a)
    a.columns = ['id','bf']
    b = pd.DataFrame(df_b)
    b.columns = ['id','bf']

    for index, row in blk.iterrows():
      id1 = row['id1']
      id2 = row['id2']

      try:
        bf1 = a[a.id == id1].values[0][1]
        bf2 = b[b.id == id2].values[0][1]
        sim = jaccard_similarity(bf1,bf2,bf_representation=bf_representation)

        if bf_representation == 'binary':
          bf1 = np.unpackbits(bf1)
          bf2 = np.unpackbits(bf2)

          if bit_power!=1:
            bf1 = bf1 * bit_power
            bf2 = bf2 * bit_power
        
        linha = [id1,id2,bf1,bf2,sim]

        if id1 == id2:
          # labels.append(np.array([1.]))
          linha.append(1)
        else:
          # labels.append(np.array([0]))
          linha.append(0)
        
        ds.append(linha)
      except IndexError:
        pass #quando o index nao e encontrado
    #fim do for

    df = pd.DataFrame(ds)      
    df.columns = ["id1","id2","bf1","bf2","sim","label"]
    return df


def __rank_all_dataset(df_a,df_b,bf_representation='binary',bit_power=1):
  '''
    Sort the dataset according to the jaccard simillarity. This method also change
    binary bf represation. It convert the bitaary into np.array(uint8) to be used 
    as input to the classifiers
    
    - bit_power: (default=1), this method is used amplify the power of the signal in the bloomfilter

    return a dataset with ["id1","id2","bf1","bf2","sim","label"]
  '''
  
  ds = []

  for i in range(0,len(df_a)):
    for j in range(0,len(df_b)):
      id1 = df_a[i][0]
      id2 = df_b[j][0]

      bf1 = df_a[i][1]
      bf2 = df_b[j][1]

      sim = jaccard_similarity(bf1,bf2,bf_representation=bf_representation)

      if bf_representation == 'binary':

        bf1 = np.unpackbits(bf1)
        bf2 = np.unpackbits(bf2)

        if bit_power!=1:
          bf1 = bf1 * bit_power
          bf2 = bf2 * bit_power

      linha = [id1,id2,bf1,bf2,sim]
      
      if id1 == id2:
        # labels.append(np.array([1.]))
        linha.append(1)
      else:
        # labels.append(np.array([0]))
        linha.append(0)
      
      ds.append(linha)

  df = pd.DataFrame(ds)      
  df.columns = ["id1","id2","bf1","bf2","sim","label"]
  return df

def extract_sample_from_anonymized(rdf,sample_method='random1',
                                    num_of_random_sample=1,
                                    num_of_negative_example_per_sample=1
                                ):
    """
        - rdf : ranked dataframe
        - sample_mentod
            - random1 : select x (num_of_random_sample * num_of_negative_example_per_sample, default 1) records random from each non matches
            - maxq13 : select x (4 * num_of_negative_example_per_sample) records pairs. From max, quantile 1, median and quantile 3 similarity values
        - num_of_random_sample : number of random sample in random1
        - num_of_negative_example_per_sample: numper of negative example per each positive example

    """
    assert sample_method in ['random1','maxq13']
    ids = rdf.id1.unique()
    sample_df = []

    id = ids[0]

    for id in ids:
        #non matches df
        df_nm = rdf[(rdf.id1 == id) & (rdf.label != 1)] # no_match
        df_m = rdf[(rdf.id1 == id) & (rdf.label == 1)]  # match

        if len(df_m) != 0:
            sample_df.append(df_m.to_numpy()[0]) # remover
            #add no match sample 
            if sample_method == 'random1':
                unique_sim = df_nm.sim.unique()
                sample_unique_sim = np.random.choice(unique_sim,num_of_random_sample)
                for i in sample_unique_sim:
                    s = df_nm[df_nm.sim ==i].sample(num_of_negative_example_per_sample,replace=True)
                    sample_df.append(s.to_numpy()[0])

            elif sample_method == 'maxq13':

                sim_values = df_nm.sim.unique()
                q1 = get_nearest_value(sim_values, df_nm.sim.quantile(0.25))
                me = get_nearest_value(sim_values, df_nm.sim.median())
                q3 = get_nearest_value(sim_values, df_nm.sim.quantile(0.75))
                ma = get_nearest_value(sim_values, df_nm.sim.max())

                for v in [q1,me,q3,ma]:
                    sample_cand = df_nm[df_nm.sim == q3]

                    if len(sample_cand) <= num_of_negative_example_per_sample: # caso o sample_cand nao tenha o numero suficiente de amostras
                        sample_df.append(sample_cand.to_numpy()[0])
                    else: # remove um sample
                        temp = sample_cand.sample(num_of_negative_example_per_sample)
                        sample_df.append(temp.to_numpy()[0])
        else:
          #colocar exemplos negativos
          pass
          if sample_method == 'random1':
            unique_sim = df_nm.sim.unique()
            sample_unique_sim = np.random.choice(unique_sim,num_of_random_sample)
            for i in sample_unique_sim:
              s = df_nm[df_nm.sim ==i].sample(num_of_negative_example_per_sample,replace=True)
              sample_df.append(s.to_numpy()[0])
          elif sample_method == 'maxq13':
            sim_values = df_nm.sim.unique()
            q1 = get_nearest_value(sim_values, df_nm.sim.quantile(0.25))
            me = get_nearest_value(sim_values, df_nm.sim.median())
            q3 = get_nearest_value(sim_values, df_nm.sim.quantile(0.75))
            ma = get_nearest_value(sim_values, df_nm.sim.max())

            for v in [q1,me,q3,ma]:
              sample_cand = df_nm[df_nm.sim == q3]
              if len(sample_cand) <= num_of_negative_example_per_sample: # caso o sample_cand nao tenha o numero suficiente de amostras
                sample_df.append(sample_cand.to_numpy()[0])
              else: # remove um sample
                temp = sample_cand.sample(num_of_negative_example_per_sample)
                sample_df.append(temp.to_numpy()[0])
  
    sample_df = pd.DataFrame(sample_df)
    sample_df.columns = ['id1','id2','bf1','bf2','sim','label']
    return sample_df


def fill_all_posbf(ds,len_max):
    """
      Fill all pos bf to the maximum length
    """
    s = np.asarray(ds)
    l = []
    r = []
    label = []

    for i in range(0,len(s)):    
        bf1 = fill_pos_bf(s[i][2],len_max)
        bf2 = fill_pos_bf(s[i][3],len_max)
        l.append(bf1)
        r.append(bf2)
        label.append(s[i][5])
        
    x = [np.asarray(l),np.asarray(r)]
    y =  np.asarray(ds.label)

    return x,y

def get_max_length_posbf(d1:np.array):
    """
     This method return the number of element of the larger pos bf
     - d1 (np.array) with the bloom filters
    """
    len_max = 0
    for i in range(0,len(d1)):
      if len(d1[i]) > len_max:
        len_max = len(d1[i])
    
    return len_max


import xxhash
import nltk
import bitarray
import numpy as np

from ds_util import generate_ngrams


def create_bf(word,bf_len,k,bf_representation='binary',num_ngrams=2):
    """
        For a given string (representing a record), this method creates a new bloom filter.

        word : string to be encoded
        bf_len : bloom filter length
        k : number of hash function
        
        bf_representation : the re
            - binary (default) : the regular bf, a bitarray with 0 and 1's
            - pos1 : this represatation return a a integer list with the positions of ones in the filter

        It is important to mention that:
            - pos1 representation might generate filter of diferent lenght, requiring a regularization
    """
    assert bf_representation in ['binary', 'pos1']

    # gerando os ngrams
    ngrams = generate_ngrams(word,ngrams=num_ngrams)
    bf = []

    #filtro vazio
    if bf_representation == 'binary':
        bf = bitarray.bitarray(bf_len)
        bf.setall(0)

        for n_gram in ngrams:
            for seed in range(k):
                pos=xxhash.xxh64(n_gram, seed=seed).intdigest() % bf_len
                bf[pos] = 1
        # end binary bf
    elif bf_representation == 'pos1':
        for n_gram in ngrams:
            for seed in range(k):
                pos=xxhash.xxh64(n_gram, seed=seed).intdigest() % bf_len
                bf.append(pos)
        
        bf = np.unique(np.asarray(bf,dtype=np.uint16))
        # end pos1 bf

    return bf


def jaccard_similarity(filter1, filter2, bf_representation='binary'):

    """
        Calculates the jaccard index between 2 bloom filters

        - filter1 : bitarray.bitarray or list depending on the bf representation
        - filter2 : bitarray.bitarray or list depending on the bf representation

        return : number between 0 and 1
    """

    # Brasileiro, i've coded this part using this website [1]
    # Please check this, chunk of code. It may contain error :p
    # 1 -http://blog.demofox.org/2015/02/08/estimating-set-membership-with-a-bloom-filter/

    assert bf_representation in ['binary', 'pos1']

    if bf_representation=='binary':
        assert type(filter1) == bitarray.bitarray
        assert type(filter2) == bitarray.bitarray 

        intersection = (filter1 & filter2).count(True)
        union = (filter1 | filter2).count(True)
    elif bf_representation=='pos1':
        assert type(filter1) == np.ndarray
        assert type(filter2) == np.ndarray
        intersection = len(list(set(filter1).intersection(filter2)))
        union = (len(set(filter1)) + len(set(filter2))) - intersection

    
    if union == 0:
        return 0

    return float(intersection) / union


def fill_pos_bf(bf,new_bf_length):
    """
        To this representation be used in ML aplication it should have the same length.
        This method fills the pos bf with zeros at the end 
    """
    try:
        z = np.zeros(new_bf_length-len(bf))
    except ValueError:
        print(len(bf)) #corrigir aqui quado for menor
        print(new_bf_length)
    return np.concatenate((bf,z))


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="phd_dataset_util",
    version="0.2.6",
    author="Thiago Nóbrega",
    author_email="thiagonobrega@gmail.com",
    description="A package that aims to execute the basic data processing from my phd experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thiaonobrega/ml_ds_util",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ),
)


import numpy as np
import pandas as pd
import nltk

def generate_ngrams(word,ngrams=2,pad=False):
    """
    Generate (plain ngrams) of a string (word)
    """
    if pad:
        # removi o str(word)
        grams =  nltk.ngrams(word, ngrams,pad_left=True, pad_right=True,left_pad_symbol='$',right_pad_symbol='$')
    else:
        grams = nltk.ngrams(word, ngrams)
    saida = []

    # try:
    for gram in grams:
        saida.append(''.join(gram))
    # except TypeError:
    #     print(type(word))
    #     print(word)
        # print(list(grams))
        # print(saida)

  
    return saida

def pre_process_raw(df,atts,ngrams=2,pad=False,return_type='qid'):
    """
        Pre-processing raw data: generate the bi-grams to be anonymized
        - remove spaces
        return_type : qid or ngrams

        return: the preprocessed dataset the mand and maximum number (maior) of ngram in the dataset
    """
    assert return_type in ['ngrams','qid']

    ldf = []
    maior = 0
    soma_media = 0
    count = 0

    for row in df.iterrows():
        id = row[1][0]
        qid = ''
        for i in range(1,len(atts)):
            try:
                for w in row[1][i].split(): #removendo espacos
                    # qid += str(w)
                    qid += w
            except AttributeError:
                qid += str(row[1][i])
        
        #
        # qid = str(qid)
        # qid = qid[:-1] # recupera ate o ultimo espaco
        
        n_grams = generate_ngrams(qid,ngrams=ngrams,pad=pad)
        soma_media += len(n_grams)
        count += 1

        if len(n_grams) >  maior:
            maior = len(n_grams)

        if return_type == 'ngrams':
            ldf.append([id,n_grams])
        else:
            ldf.append([id,qid])

    media_ngram = int(soma_media/count)
    return ldf, media_ngram, maior

def extract_sample(df1,sample_size,duplicate_rate=0.1,
                    df2=[],
                    return_raw_sample=False,
                    return_type='qid',
                    atts='all'
                ):
    """
        Extract a sample from a dataset (or a pair of dataset, if df2 parameter is set)
        - sample_size : number of records
        - duplicate_rate: percent of duplicated (0...1)
        - df2: the second dataset to be extracted
        - return_raw_sample (default False): return the raw sample, for debug.
        - atts

        return df_a,df_b,mean_ngram,max_ngram ( raw_dfa, raw_dfb )
    """
    assert duplicate_rate <= 1
    assert duplicate_rate > 0

    df2_flag = False
    if len(df2) > 0:
        df2_flag = True

    num_of_duplicated_records_df_b = int(sample_size * duplicate_rate)
    num_of_no_duplicated_records_df_b = sample_size - num_of_duplicated_records_df_b
    
    df_a = df1.sample(sample_size)
    if df2_flag == False:
        df_b = pd.concat([df1.sample(num_of_no_duplicated_records_df_b),
                            df_a.sample(num_of_duplicated_records_df_b) #corrigido
                        ])
    else:
        df_b = pd.concat([df2.sample(num_of_no_duplicated_records_df_b),
                            df_a.sample(num_of_duplicated_records_df_b)
                        ])

    #seleciona attributos
    if type(atts) != list:
        atts = list(df_a.columns)
    else:
        #filter attributes
        df_a = df_a[atts]
        df_b = df_b[atts]

    #preprocessa os dados
    df_a_proc , df_a_mean_ngram, df_a_max_ngram = pre_process_raw(df_a,atts,return_type=return_type)
    df_b_proc , df_b_mean_ngram, df_b_max_ngram = pre_process_raw(df_b,atts,return_type=return_type)
    #max and mean number of ngram in the sample
    max_ngram = max(df_a_max_ngram,df_a_max_ngram)
    mean_ngram = int( (df_a_mean_ngram+df_b_mean_ngram)/2 )

    if return_raw_sample:
        return df_a_proc,df_b_proc,mean_ngram,max_ngram,df_a,df_b
    else:
        return df_a_proc,df_b_proc,mean_ngram,max_ngram

